/*
m_kquantise - quantise a control signal to a predefined grid

DESCRIPTION
This opcode will quantise a control signal to a certain grdi or stepsize.
If you set a step size of 1 and apply the control signal to a frequency, you
will get a harmonic series.

SYNTAX
kout m_kquantise ksignal, kmode, kstep, kquality, kparam, ifn_scales

INITIALIZATION
ifn_scales - Function table holding the numbers of function tables, which
	hold the grade multipliers for the quantisation scales.
	Each scale function table holds 12 values, for a major scale this could be:
	1 2^(2/12) 2^(2/12) 2^(4/12) 2^(5/12) 2^(5/12) 2^(7/12) 2^(7/12) 2^(9/12)
	2^(9/12) 2^(11/12) 2

PERFORMANCE
kout - quantised output signal
ktrig - Trigger signal, active whenever a new quantisation step is reached
ksignal - Input control signal
kmode - Mode of the quantiser
	0: off
	1: linear (quantise to each step)
	2: multiply (quantise to each kparam * kstep)
	3: Scale (quantise to specified scale)
kstep - Step size to quantise to (or number of the scale to quantise to in
	scale mode)
kquality - Quality of the quantisation process, between 0 and 1.
kparam - Parameter of the quantisation
	in multiply and scale mode the base frequency in Hz

CREDITS
Author: Jeanette C., with the help of the Csound mailing list, especially
Stankevich.
*/

opcode m_kquantise, kk, kkkkki
	ksig, kmode, kstep, kquality, kparam, ifn_scales xin

	kquality limit kquality, 0, 1
	if kmode == 0 kgoto Bypass
	if kmode == 1 kgoto Linear
	if kmode == 2 kgoto Multiply
	if kmode == 3 kgoto Scale

	; Bypass mode, copy input to output
Bypass:
	kout = ksig
	ktrig = 0
	kgoto Output

	; Linear mode
Linear:
	kout = ksig + (round(ksig/kstep) * kstep - ksig) * kquality
	ktrig changed kout
	kgoto Output

	; Multiply mode
Multiply:
	kquant = kstep * kparam
	kout = ksig + (round(ksig/kquant) * kquant - ksig) * kquality
	ktrig changed kout
	kgoto Output

	; Scale mode
Scale:
	if changed(ksig) == 1 then ; only process if ksig changes
		; Calculate the octave and 12-tone equal-tempered note
		krelativenote = log2(ksig / kparam)
		koctave = 2^(int(krelativenote))
		knote = round(frac(krelativenote) * 12)
		kscalefn tablei kstep, ifn_scales, 0, 0, 1
		krealnote tablekt knote, kscalefn, 0
		kout = kparam * koctave * krealnote
	endif
	ktrig changed kout

	; Output the signal and trigger
Output:
	xout kout, ktrig
endop
